/*-----------------------------------------------------------------------------
     swp2ga.c: to read swp file and to write 4-byte binary file for GrADS

                                       1999/09/17  Ver 1.0 by Shin Satoh

 * ver 0.1: output array alloc at first
 * ver 0.2: memcpy (dynamic memory allocation)
 * ver 0.3: fixed output dimension of jmax
 * ver 0.4: output spol and bist data separately (swp2ga_spol, swp2ga_bist)
 * ver 0.5: input commands
 * ver 1.0: simple printout, select params, 
            output GrADS data descriptor file (*.ctl)
  -----------------------------------------------------------------------------*/
/*
 * This file contains the following routines
 *
 * main
 * dd_absorb_header_info
 * dd_absorb_ray_info
 * dd_absorb_rotang_info
 * dd_absorb_seds
 * dd_return_sweepfile_structs_v3
 * 
 * dd_return_difs_ptr       from dda_common.c
 * dd_next_comment          from dda_common.c
 * dd_hrd16_uncompressx     from dorade_share.c
 * dd_hrd16LE_uncompressx   from dorade_share.c
 * dts_print                from dorade_share.c
 * str_terminate            from dorade_share.c
 * dorade_time_stamp        from dorade_share.c
 *
 */

# include "Volume.h"
# include "Comment.h"
# include "RadarDesc.h"
# include "Correction.h"
# include "Parameter.h"
# include "CellVector.h"
# include "CellSpacing.h"
# include "Sweep.h"
# include "Platform.h"
# include "Ray.h"
# include "Pdata.h"
# include "Qdata.h"
# include "FieldParam.h"
# include "super_SWIB.h"

# include "dd_defines.h"
# include "dorade_share.h"
# include "dd_general_info.h"

# include "dd_files.h"
# include "input_limits.h"
# include "input_sweepfiles.h"
# include "dd_stats.h"
# include "function_decl.h"
# include "dgi_func_decl.h"
# include <stdlib.h>
# include <string.h>

static struct dd_input_sweepfiles_v3 *dis_v3=NULL;
static struct dd_comm_info *first_comment=NULL;
static struct dd_input_filters *keep_difs=NULL;

int dd_absorb_rotang_info();
int dd_absorb_seds();
void ddswp_stuff_ray();
void ddswp_stuff_sweep();
void dd_update_ray_info();
void dd_update_sweep_info();
int ddswp_new_sweep_v3();

/* LittleEndian for Linux */
# ifdef LITTLE_ENDIAN   /* This is usually defined in /usr/include/endian.h */
int LittleEndian=YES;
# else
int LittleEndian=NO;
# endif

/*---------------------------------------------------------------------------------*/
# define NO_DATA -999.0

int main(int argc, char *argv[])
{
    DGI_PTR dgi;
    DDS_PTR dds;
    DIS_PTR dis, dd_return_sweepfile_structs_v3();

    FILE *fpg, *fpt;
    static char fin[256],fout[256],fout_dir[256],ftxt[256];
    static char fname[64],strp[32],strq[32];
    float *sptr;                    /*  float sdat[18][408+1][534+3]; */
    float *secptr,*azptr,*elptr;    /*  float daysec[408],az[408],el[408]; */
    short *qptr;
    float rcp_scale, bias;
    float xmin,ymin,zmin,dx,dy,dz,undef;
    int   icf=0, ctl_out_flag=0, verbose_flag=0, txt_out_flag=0; 
    int   nfin,nfin_max=0,fin_argv=0,fin_len,fout_flag=0,fout_len;
    int   np_out[MAX_PARMS];
    int   np,np_max=0,np_argv=0,np_act=0;
    int   i,j,k,ii,jj,kk,ijk,return_bc;
    int   imax,jmax,jmax_act,jmax_fix=0,kmax;
    float elmean=0.0;

/*===== Check input commands =====*/
    for (i=0; i<argc; i++){

      icf=0;
      if(strcmp(argv[i],"-i")==0) icf=1;
      if(strcmp(argv[i],"-o")==0) icf=2;
      if(strcmp(argv[i],"-p")==0) icf=3;
      if(strcmp(argv[i],"-f")==0) icf=4;
      if(strcmp(argv[i],"-c")==0) icf=11;
      if(strcmp(argv[i],"-v")==0) icf=12;
      if(strcmp(argv[i],"-t")==0) icf=13;
      if(strcmp(argv[i],"-h")==0) icf=51;
      if(strcmp(argv[i],"-help")==0) icf=51;
      if(i==0 && argc>2) icf=49; /* nothing to do */
      if(i==0 && argc<3) icf=90; /* no input command */

      switch(icf){
        case 1:   /* [-i] input file name */
 	   while(1){
  	     i++;    
             if(i>=argc){
               if(nfin_max==0) icf=91; 
               break;
             }
             if(strncmp(argv[i],"-",1)==0){
               i--;
               if(nfin_max==0) icf=91;  /* case of no input file */
               break;
	     }else{
               if(fin_argv==0) fin_argv = i;
               nfin_max++;
             }
	   }
           break;

        case 2:   /* [-o] output file name */
             i++;   
             if(i>=argc){
               icf=82; 
               break;
             }
             if(strncmp(argv[i],"-",1)==0){       
               i--;
               icf=82;                 /* case of no output file */
             }else{
               strcpy(fout,argv[i]);
               fout_flag = 1;
               fout_len = strlen(fout) - 1;   
               if(strncmp(&fout[fout_len],"/",1)==0) {  /* case of directory */
                 strcpy(fout_dir,fout);
                 fout_flag = 2;
               }
             }
           break;

        case 3:   /* [-p] output parameter name */
 	   while(1){
  	     i++;    
             if(i>=argc){
               if(np_max==0) icf=83; 
               break;
             }
             if(strncmp(argv[i],"-",1)==0){
               i--;
               if(np_max==0) icf=83;  /* case of no parameter name */
               break;
	     }else{
               if(np_argv==0) np_argv = i;
               np_max++;
             }
	   }
           break;

        case 4:  /* [-f] fixed num of rays */
             i++;    
             if(i>=argc){
               icf=84; 
               break;
             }
             if(strncmp(argv[i],"-",1)!=0){
               jmax_fix = atoi(argv[i]);
             }else{
               i--;
               icf=84;             /* case of no fixed num of rays */
             }
             break;

        case 11:
           ctl_out_flag = 1;
           break;
        case 12:
           verbose_flag = 1;
           break;
        case 13:
           txt_out_flag = 1;
           break;

        default: 
          break;
      }

      /* Warning */
      if(icf==82) printf("# WARNING: no output file ---> name automatically\n");
      if(icf==83) printf("# WARNING: no parameter ---> output all parameters\n");
      if(icf==84) printf("# WARNING: no fixed num of rays ---> output variable dim.\n");

      /* Error */
      if(icf==0)  printf("# ERROR: Invalid command: %s\n\n",argv[i]);
      if(icf==91) printf("# ERROR: no input file\n\n");
      if(icf>=90 || icf==51 || icf==0) {
          printf("usage: %s\n",argv[0]);
          printf("      -i [input file(s)]       : possible to use wild card (ex. swp.*)\n");
          printf("     (-o [output path/file]    : if not, name automatically (gart.))\n");
          printf("     (-p [output parameter(s)] : (ex. -p DM DZ VR )\n");
          printf("     (-f [fixed num of rays]   : fixed output dimension (Num of AZ data)\n");
          printf("     (-c                       : output GrADS ctl file )\n");
          printf("     (-v                       : be verbose )\n");
          printf("     (-t                       : output text-form file )\n");
          printf("     (-h -help                 : show this usage )\n");
          if (icf==51) exit(0);
          exit(1);
      }

    }   /* end of for(i=0; i<argc; i++) */

/*===== Warning for input commands =====*/
    if(nfin_max>1 && fout_flag==1){
      printf("# WARNING: Ignore the output file name: %s,",fout);
      printf(" because plural input files were appointed\n"); 
      fout_flag = 0;
    }

/*===== Main loop for input files ====================================================*/
    for(nfin=0; nfin<nfin_max; nfin++){

/*===== Make input & output file name =====*/
      i = fin_argv + nfin;
      strcpy(fin,argv[i]);

      strcpy(fname,fin);
      fin_len = strlen(fin) - 1;
      for(i=0; i<fin_len; i++){
        if(strncmp(&fin[i],"/",1)==0) strcpy(fname,&fin[i+1]);
      }

      switch(fout_flag){
      case 0:  /* no fout */
        if(strncmp(fname,"swp.",4)==0){
          strcpy(fout,"gart.");
          strcat(fout,&fname[4]);
        }else if(strncmp(fname,"swp_",4)==0){
          strcpy(fout,"gart_");
          strcat(fout,&fname[4]);
        }else{
          strcpy(fout,"gart.");
          strcat(fout,&fname[0]);
        }
        break;
      case 1:  /* fout is a file name */
        break;
      case 2:  /* fout is a path name */
        if(strncmp(fname,"swp.",4)==0){
          strcpy(fout,fout_dir);
          strcat(fout,"gart.");
          strcat(fout,&fname[4]);
        }else if(strncmp(fname,"swp_",4)==0){
          strcpy(fout,fout_dir);
          strcat(fout,"gart_");
          strcat(fout,&fname[4]);
        }else{
          strcpy(fout,fout_dir);
          strcat(fout,"gart.");
          strcat(fout,&fname[0]);
        }
        break;
      default:
        break;
      }

      if(txt_out_flag){
	switch(fout_flag){
	case 0:  /* no fout */
	  if(strncmp(fname,"swp.",4)==0){
	    strcpy(ftxt,"txt.");
	    strcat(ftxt,&fname[4]);
	  }else if(strncmp(fname,"swp_",4)==0){
	    strcpy(ftxt,"txt_");
	    strcat(ftxt,&fname[4]);
	  }else{
	    strcpy(ftxt,"txt.");
	    strcat(ftxt,&fname[0]);
	  }
	  break;
	case 1:  /* fout is a file name */
	case 2:  /* fout is a path name */
	  if(strncmp(fname,"swp.",4)==0){
	    strcpy(ftxt,fout_dir);
	    strcat(ftxt,"txt.");
	    strcat(ftxt,&fname[4]);
	  }else if(strncmp(fname,"swp_",4)==0){
	    strcpy(ftxt,fout_dir);
	    strcat(ftxt,"txt_");
	    strcat(ftxt,&fname[4]);
	  }else{
	    strcpy(ftxt,fout_dir);
	    strcat(ftxt,"txt.");
	    strcat(ftxt,&fname[0]);
	  }
	  break;
	default:
	  break;
	}
      }

      printf("No=%3d input file : %s\n",nfin+1,fin);
      printf("       output file: %s\n",fout);
      if(txt_out_flag==1) printf("       text output file: %s\n",ftxt);

      if(verbose_flag==1){
        printf("fixed output number of rays(AZ data in PPI): %d\n",jmax_fix);
        if(ctl_out_flag==1) printf("output GrADS ctl file\n");
	if(txt_out_flag==1) printf("output text file\n");
        if(verbose_flag==1) printf("verbose mode\n");
      }
 
/*===== Allocate major struct =====*/
    if(nfin==0){
      dgi = (struct dd_general_info *)malloc(sizeof(struct dd_general_info));
    }
    if(nfin==0){
      dgi->dds = (struct dds_structs *)malloc(sizeof(struct dds_structs));
    }
    if(nfin==0){
      dgi->dds->dts = (struct d_time_struct *)malloc(sizeof(struct d_time_struct));
      dds=dgi->dds;
    }
    dis = dd_return_sweepfile_structs_v3();

/*===== Open file =====*/
    if((dgi->in_swp_fid = open(fin, 0)) < 0 ){
	printf("File not open ! : %s \n",fin); exit(1); }

    if((fpg = fopen(fout,"w" )) == NULL ){
        printf("File not open ! : %s \n",fout); exit(1);   }

    if(txt_out_flag==1){
      if((fpt = fopen(ftxt,"wt" )) == NULL ){
        printf("File not open ! : %s \n",ftxt); exit(1);   }
    }

/*===== Read header =====*/
    dd_absorb_header_info(dgi,dis);

    /* Set dis->usi_top */
    fin_len = strlen(fin) - 1; ii=0;
    for(i=0; i<fin_len; i++){
       if(strncmp(&fin[i],"/",1)==0) ii=i+1;
    }
    if(ii!=0) {
      strncpy(dis->usi_top->directory,&fin[0],ii);
    }
    strcpy(dis->usi_top->filename,&fin[ii]);     
    if(strcmp(dis->usi_top->radar_name,"")==0){
      strncpy(dis->usi_top->radar_name,dds->radd->radar_name, 8);
    }

    /* Print header info and parameters */
    if(verbose_flag==1){
      print_header_info(dgi,dis);
      print_params_info(dgi);
    }

    if(txt_out_flag==1){
      text_output_header_info(dgi,dis,fpt);
    }

    /* Allocate store data region */
    imax = dds->celv->number_cells;
    if(jmax_fix==0){
      jmax = dds->swib->num_rays;
    }else{    
      jmax = jmax_fix;
    }
    kmax = dgi->source_num_parms;
    if(kmax>MAX_PARMS){ 
      printf("# ERROR: dimension over of MAX_PARMS\n"); 
      exit(1);
    } 
    secptr = (float *)malloc(jmax*sizeof(float));
    azptr  = (float *)malloc(jmax*sizeof(float));
    elptr  = (float *)malloc(jmax*sizeof(float));
    sptr = (float *)malloc(kmax*(imax+3)*(jmax+1)*sizeof(float));
    if(!sptr){
      printf("# ERROR: Cannot alloc memory, maybe too much dimension");
      exit(1);
    }
    if(verbose_flag==1) printf("======== END of the header block (read RYIB) ==========\n\n");

/*===== Read ray_info & data =====*/
    jj=0;
    while(1){
      return_bc = dd_absorb_ray_info(dgi);
      if(return_bc == 1) break;
# ifdef DEBUG
      printf("====== End of dd_absorb_ray_info:  binary_count=%d\n",return_bc);
# endif

      /* Store data */
      if(jj<jmax){
         for(k=0; k < dgi->source_num_parms; k++) {
            qptr = (short *)dds->qdat_ptrs[k];
            rcp_scale = 1./dds->parm[k]->parameter_scale;
            bias = dds->parm[k]->parameter_bias;
            for(ii=0; ii<dds->celv->number_cells; ii++){
               j = jj+1;
               i = ii+3;
               ijk=(k*(jmax+1)*(imax+3))+(j*(imax+3))+i;
               if(*(qptr+ii) != (short)dds->parm[k]->bad_data){ 
                  *(sptr+ijk) = DD_UNSCALE((float)*(qptr+ii), rcp_scale, bias);
                         /* sdat[k][j][i] = DD_UNSCALE((float)*(qptr+ii), rcp_scale, bias); */
               }else{
                  *(sptr+ijk) = NO_DATA;  
                         /* sdat[k][j][i] = NO_DATA; */
	       }
            }
# ifdef DEBUG
            str_terminate(strp, dds->parm[k]->parameter_name, 8);
            printf("#pn=%2d [%s]:  sptr points %10d\n",k,strp,sptr+ijk);
# endif
          }
          *(secptr+jj) = dds->ryib->hour * 3600. +
                         dds->ryib->minute * 60. +
                         dds->ryib->second       +
                         dds->ryib->millisecond * 0.001;
          *(azptr+jj)  = dds->ryib->azimuth;    /* az[jj] = dds->ryib->azimuth; */
          *(elptr+jj)  = dds->ryib->elevation;  /* el[jj] = dds->ryib->elevation; */
      } /* end of if(jj<jmax) */

      /* Mean Elevation */
      elmean = elmean + dds->ryib->elevation;

      /* Print time and angles */
      if(verbose_flag==1){
        if(jj<10 || jj>(dds->swib->num_rays-11)){
          if(jj>=jmax) printf("# Abondoned data: ");
          printf("RAY_NUM=%3.3d: DATE=%4d/%2.2d/%2.2d ",  
                 jj+1,dds->vold->year,dds->vold->month,dds->vold->day);
          printf("(%d julian_day)  TIME=%2.2d:%2.2d:%2.2d.%3.3d  ",  
                 dds->ryib->julian_day,dds->ryib->hour,dds->ryib->minute,
                 dds->ryib->second,dds->ryib->millisecond);
          printf("AZ=%7.3f  EL=%6.3f\n",
                 dds->ryib->azimuth,dds->ryib->elevation);
        }
        if(jj==10) printf("     -----<omit>-----\n");
      }else{
        if(jj==0){
            printf("       start time and angles: %4d/%2.2d/%2.2d,%2.2d:%2.2d:%2.2d  ",
            dds->vold->year,dds->vold->month,dds->vold->day,
            dds->ryib->hour,dds->ryib->minute,dds->ryib->second);
            printf("AZ=%5.1f  EL=%4.1f\n",
                 dds->ryib->azimuth,dds->ryib->elevation);
        }
        if(jj==dds->swib->num_rays-1){
            printf("       end time and angles:   %4d/%2.2d/%2.2d,%2.2d:%2.2d:%2.2d  ",
            dds->vold->year,dds->vold->month,dds->vold->day,
            dds->ryib->hour,dds->ryib->minute,dds->ryib->second);
            printf("AZ=%5.1f  EL=%4.1f\n",
                 dds->ryib->azimuth,dds->ryib->elevation);
        }
      }

      jj++;

    }  /* end of while(1) */

    /* Mean Elevation */
    elmean = elmean/((float)jj);

    if(verbose_flag==1) printf("======== END of the data block (read NULL) ==========\n\n");

/*===== Check jmax and output parameter =====*/
    /* check jmax */
    jmax_act = dgi->source_ray_num;
    if(verbose_flag==1){
      printf("  Num. of rays: header info jmax (SWIB)=%d\n",dds->swib->num_rays); 
      if(jmax_fix!=0)
      printf("                fixed jmax by force    =%d\n",jmax_fix); 
      printf("                counting actual jmax   =%d\n\n",jmax_act);
    }
    if(jmax_act > jmax && jmax==dds->swib->num_rays){
      printf("# WARNING: actual-jmax(%d) larger than header-jmax(%d)",jmax_act,jmax);
      printf(" ---> abandon data larger than header-jmax\n");
    }
    if(jmax_act > jmax && jmax==jmax_fix){
      printf("# WARNING: actual-jmax(%d) larger than fixed-jmax(%d)",jmax_act,jmax);
      printf(" ---> abandon data larger than fixed-jmax\n");
    }

    /* print dimension in the input data */
    printf("       included data dims:  %d(params) * %d(rays) * %d(cells)\n",
           kmax,jmax_act,imax);

    /* print included parameter name */
    if(np_max > 0){
      printf("       included parameters: ");
      for(k=0; k<kmax; k++){
        str_terminate(strp, dds->parm[k]->parameter_name, 8);
        printf("%s ",strp);
      }
      printf("\n");
    }

    /* check and print output parameter */
    np_act=0;
    printf("       output parameters:   ");
    if(np_max > 0){
      for(k=0; k<kmax; k++){
        str_terminate(strp, dds->parm[k]->parameter_name, 8);
        for (np=0; np<np_max; np++){
          i = np_argv + np;
          str_terminate(strq,argv[i],8);
          if (strcmp(strp,strq)==0){
            np_out[np_act]=k;
            np_act++;
            printf("%s ",strp);
          }
        }
      }
    }else{
      np_act=kmax;
      for(k=0; k<kmax; k++){
        np_out[k]=k;
        str_terminate(strp, dds->parm[k]->parameter_name, 8);
        printf("%s ",strp);
      }
    }      
    printf(" (%d params)\n",np_act);

    if(np_max>0 && np_act!=np_max){
      printf("# WARNING: incorrect parameters may be specified by -p ...\n"); 
    }

# ifdef DEBUG
      printf("output param no.= ");
      for(np=0; np<np_act; np++){printf("%d, ",np_out[np]);}
      printf("... total=%d\n",np_act);
# endif

/*===== Output 4-byte data =====*/
    /* Fill dummy data over actual dimension jmax */
    for(j=jmax_act; j<jmax+1; j++){
      *(secptr+j-1) = NO_DATA;
      *(azptr+j-1)  = NO_DATA;
      *(elptr+j-1)  = NO_DATA;
    }
    for(k=0; k<kmax; k++){
    for(j=jmax_act; j<jmax+1; j++){
    for(i=3; i<imax+3; i++){
       ijk=(k*(jmax+1)*(imax+3))+(j*(imax+3))+i;
       *(sptr+ijk)   = NO_DATA;
    }}}

    /* Fill distance data on each cell */
    for(k=0; k<kmax; k++){
    for(i=3; i<imax+3; i++){
      j=0;  ijk=(k*(jmax+1)*(imax+3))+(j*(imax+3))+i;
      *(sptr+ijk) = (float)dds->celv->dist_cells[i-3];  /* sdat[k][0][i]=(float)dds->celv->dist_cells[i-3]; */
    }}
    
    /* Fill time, AZ, and EL data on each ray */
    for(k=0; k<kmax; k++){
    for(j=1; j<jmax+1; j++){
      i=0;  ijk=(k*(jmax+1)*(imax+3))+(j*(imax+3))+i;
      *(sptr+ijk)   = *(secptr+j-1);    /* sdat[k][j][0]=(float)daysec[j-1]; */
      *(sptr+ijk+1) = *(azptr+j-1);     /* sdat[k][j][1]=(float)az[j-1]; */
      *(sptr+ijk+2) = *(elptr+j-1);     /* sdat[k][j][2]=(float)el[j-1]; */
    }}

    /* Fill dummy data within artificial arrys */
    for(k=0; k<kmax; k++){
      i=0; j=0; ijk=(k*(jmax+1)*(imax+3))+(j*(imax+3))+i;
      *(sptr+ijk)   = NO_DATA;          /* sdat[k][0][0]=NO_DATA; */
      *(sptr+ijk+1) = NO_DATA;          /* sdat[k][0][1]=NO_DATA; */
      *(sptr+ijk+2) = NO_DATA;          /* sdat[k][0][2]=NO_DATA; */
    }

    /* Write restrict output parameters to disk */
    for(kk=0; kk<kmax; kk++){
      k=-99;
      for (np=0; np<np_act; np++){
         if(np_out[np]==kk) k=kk;
      }
      if(k>=0){
         i=0; j=0;
         ijk=(k*(jmax+1)*(imax+3))+(j*(imax+3))+i;
         fwrite((sptr+ijk),1,(jmax+1)*(imax+3)*sizeof(float),fpg);
      }
    }
    fclose (fpg);

    /* Output GrADS control file */
    if(ctl_out_flag==1){
      xmin=(float)dds->celv->dist_cells[0];
      dx=(float)dds->celv->dist_cells[1]-xmin;
      xmin=xmin*0.001; dx=dx*0.001;  /* [m] -> [km] */
      xmin=xmin-(dx*3);              /* x=1:sec, x=2:az, y=3:el */
      ymin=*(azptr);              dy=*(azptr+1)-ymin;
      ymin=ymin-dy;                  /* y=1:distance[m] */
      zmin=*(elptr+jmax_act-2);   dz=zmin-*(elptr);
      if(dz==0.0) dz=1.0;
      undef=NO_DATA;

      ggrads_ctl (dgi,dis,fout,np_out,np_act,imax+3,jmax+1,1,
                  xmin,ymin,zmin,dx,dy,dz,undef);
    }

    /* Output text-form file */
    if(txt_out_flag==1){
      text_output_sweep_info(elmean,dds,fpt);
      text_output_data(imax,jmax,kmax,sptr,dds,fpt);
    }

    /* Print output params and size */
    printf("       output data size:    %d * %d * %d * 4 byte\n",
           np_act,(jmax+1),(imax+3));
    if(verbose_flag==1) printf("\n\n\n\n");

/* ===== Free malloc of store data region ===== */
    free(sptr);
    free(secptr);
    free(azptr);
    free(elptr);

}  /* end of main loop for input files: for(nfin=0; nfin<nfin_max; nfin++) */ 

    return(0);
}
/* c------------------------------------------------------------------------ */

int dd_absorb_header_info(dgi,dis)
    DGI_PTR dgi;
    DIS_PTR dis;
{
    int ii, jj, pn, bc=0, loop_count=0, gottaSwap=NO, broken=NO;
    int gotta_cfac = NO, gotta_celv = NO;
    char *aa, str[256], *str_terminate(), *getenv(), radar_name[12];
    char *deep6=0;
    struct generic_descriptor *gd;
    DDS_PTR dds=dgi->dds;
    DIS_PTR dd_return_sweepfile_structs_v3();
    /* struct dd_input_sweepfiles_v3 *dis, *dd_return_sweepfile_structs_v3(); */
    unsigned long gdsos;

    /*
     * absorb header info
     */
    dis = dd_return_sweepfile_structs_v3();
    dgi_buf_rewind(dgi);

    if((dgi->buf_bytes_left=read(dgi->in_swp_fid, dgi->in_buf, MAX_READ ))
       < 1 ) {
	printf("Bad read or eof state: %d in dd_absorb_header\n"
	       , dgi->buf_bytes_left );
	exit(1);
    }
    dis->MB += BYTES_TO_MB(dgi->buf_bytes_left);
    dis->rec_count++;
    dis->sweep_count++;
    dis->file_count++;
    dgi->num_parms = 0;
# ifdef DEBUG
            printf("+++++++ Read record count=%d (%f MB) +++++++\n",
            dis->rec_count,dis->MB);
# endif

    for(;;) {
	loop_count++;
	gd = (struct generic_descriptor *)dgi->in_next_block;
	gdsos = (unsigned long)gd->sizeof_struct;
        /* If gdsos is greater than 65500 byte, try to change 
           byte-order (swack4) */
	if(gdsos > MAX_REC_SIZE) {
	   gottaSwap = YES;
	   swack4(&gd->sizeof_struct, &gdsos);
	}
	if(gdsos < sizeof(struct generic_descriptor) ||
	   gdsos > MAX_REC_SIZE) {
		 broken = YES;
		 break;
	}
# ifdef DEBUG
            printf("### loop count=%2d  sizeof_struct(gdsos)=%u\n",loop_count,gdsos);
# endif

	if(strncmp(dgi->in_next_block, "RYIB", 4) == 0 ){
	      break;
	}

	if(strncmp(dgi->in_next_block, "SSWB", 4) == 0 ) {
	  if(!dgi->dds->sswb){
              dgi->dds->sswb = (struct super_SWIB *)malloc(sizeof(struct super_SWIB));
# ifdef DEBUG
              printf("#--> Find SSWB: alloc size=%d\n",sizeof(struct super_SWIB));
# endif
          }
	  if(gottaSwap) {
	      if(LittleEndian){
		 ddin_crack_sswbLE
		   (dgi->in_next_block, (char *)dds->sswb, (int)0);
	      }else{
		 ddin_crack_sswb
		   (dgi->in_next_block, (char *)dds->sswb, (int)0);
	      }
	  }else{
	      memcpy((char *)dds->sswb, dgi->in_next_block, gdsos);
	  }
	  dgi->source_num_parms = dgi->num_parms = 0;
	  bc += gdsos;
	}
	else if(strncmp(dgi->in_next_block,"VOLD",4)==0) {
          if(!dgi->dds->vold){
              dgi->dds->vold = (struct volume_d *)malloc(sizeof(struct volume_d));
# ifdef DEBUG
              printf("#--> Find VOLD: alloc size=%d\n",sizeof(struct volume_d));
# endif
          }
	  if(gottaSwap) {
	      ddin_crack_vold(dgi->in_next_block, dds->vold, (int)0);
	  }else{
	      memcpy((char *)dds->vold, dgi->in_next_block, gdsos);
	  }
	  if(dgi->source_vol_num != dds->vold->volume_num) {
	      dgi->new_vol = YES;
	      dis->vol_count++;
	  }
	  dgi->source_vol_num = dds->vold->volume_num;
	  bc += gdsos;
	}
	else if(strncmp(dgi->in_next_block,"RADD",4)==0 ) {
          if(!dgi->dds->radd){
              dgi->dds->radd = (struct radar_d *)malloc(sizeof(struct radar_d));
# ifdef DEBUG
              printf("#--> Find RADD: alloc size=%d\n",sizeof(struct radar_d));
# endif
          }
          if(!dgi->dds->asib){
              dgi->dds->asib = (struct platform_i *)malloc(sizeof(struct platform_i));
# ifdef DEBUG
              printf("          ASIB: alloc size=%d\n",sizeof(struct platform_i));
# endif
          }
	  if(gottaSwap) {
	      ddin_crack_radd(dgi->in_next_block, dds->radd, (int)0);
	      if(gdsos < sizeof(struct radar_d))
		{ dds->radd->extension_num = 0; }
	  }else{
	      memcpy((char *)dds->radd, dgi->in_next_block, gdsos);
	  }
	  strcpy(radar_name, str_terminate(str, dds->radd->radar_name, 8));
	  strcpy(dgi->radar_name, radar_name);
	  for(ii=0; ii < MAX_PARMS; dds->field_present[ii++]=NO);
	  /* always put the lat/lon/alt in the ASIB */
	  dds->asib->longitude = dds->radd->radar_longitude;
	  dds->asib->latitude = dds->radd->radar_latitude;
	  dds->asib->altitude_msl = dds->radd->radar_altitude;
	  dds->asib->heading = 0;
	  bc += gdsos;
	}
	else if(strncmp(dgi->in_next_block,"CELV",4)==0 ) {
          if(!dgi->dds->celv){
              dgi->dds->celv = (struct cell_d *)malloc(sizeof(struct cell_d));
# ifdef DEBUG
              printf("#--> Find CELV: alloc size=%d\n",sizeof(struct cell_d));
# endif
          }
          if(!dgi->dds->celvc){
              dgi->dds->celvc = (struct cell_d *)malloc(sizeof(struct cell_d));
# ifdef DEBUG
              printf("          CELVC:alloc size=%d\n",sizeof(struct cell_d));
# endif
          }
	  if(gottaSwap){
	      ddin_crack_celv(dgi->in_next_block, dds->celv, (int)0);
	      swack_long(dgi->in_next_block+12, &dds->celv->dist_cells[0]
			  ,(int)dds->celv->number_cells);
	  }else{
	      memcpy((char *)dds->celv, dgi->in_next_block, gdsos);
	  }
	  memcpy((char *)dds->celvc, dds->celv, gdsos);
	  bc += gdsos;
	  for(ii=0; ii < MAX_PARMS; ii++){
	      if(dgi->dds->field_present[ii]){
		  dd_alloc_data_field(dgi, ii);
              }
	  }
	}
	else if(strncmp(dgi->in_next_block,"CFAC",4)==0 ) {
          if(!dgi->dds->cfac){
            dgi->dds->cfac = (struct correction_d *)malloc(sizeof(struct correction_d));
# ifdef DEBUG
            printf("#--> Find CFAC: alloc size=%d\n",sizeof(struct correction_d));
# endif
          }
	  /* get CFAC stuff */
	  if(aa = getenv("CFAC_FILES")) {
		dd_absorb_cfac(aa, radar_name, dgi->dds->cfac);
		dgi->ignore_cfacs = YES;
	  }
	  if(!dgi->ignore_cfacs) {
	       if(gottaSwap) {
		  ddin_crack_cfac(dgi->in_next_block, dds->cfac, (int)0);
	       }else {
		  memcpy((char *)dds->cfac, dgi->in_next_block, gdsos);
	       }
	  }
	  dd_set_uniform_cells(dgi->dds);
	  bc += gdsos;
	  for(ii=0; ii < MAX_PARMS; ii++) {
		if(dgi->dds->field_present[ii])
		      dd_alloc_data_field(dgi, ii);
	  }
	}
	else if(strncmp(dgi->in_next_block, "SWIB", 4) == 0 ) {
           if(!dgi->dds->swib){
              dgi->dds->swib = (struct sweepinfo_d *)malloc(sizeof(struct sweepinfo_d));
# ifdef DEBUG
              printf("#--> Find SWIB: alloc size=%d\n",sizeof(struct sweepinfo_d));
# endif
           }
	   if(gottaSwap) {
	      ddin_crack_swib(dgi->in_next_block, dds->swib, (int)0);
	   }else{
	      memcpy((char *)dds->swib, dgi->in_next_block, gdsos);
	   }
	   dgi->source_num_rays = dds->swib->num_rays;
	   dgi->source_sweep_num = dds->swib->sweep_num;
	   dgi->source_ray_num = 0;
	   dgi->new_sweep = YES;
	   bc += gdsos;
	}
	else if(strncmp(dgi->in_next_block, "PARM", 4) == 0 ) {
	   pn = dgi->source_num_parms++;
	   dgi->num_parms = dgi->source_num_parms;
	   dds->field_present[pn] = YES;

	   if(!dgi->dds->parm[pn]){
               dgi->dds->parm[pn] = (struct parameter_d *)malloc(sizeof(struct parameter_d));
# ifdef DEBUG
               printf("#--> Find PARM: alloc size=%d    param_num=%2d\n",
                      sizeof(struct parameter_d),pn);
# endif
           }
	   if(gottaSwap) {
	       ddin_crack_parm(dgi->in_next_block, dds->parm[pn], (int)0);
	       if(gdsos < sizeof(struct parameter_d))
		 { dds->parm[pn]->extension_num = 0; }
	   }else {
	       memcpy((char *)dds->parm[pn], dgi->in_next_block, gdsos);
	   }
	   str_terminate(str, dds->parm[pn]->param_description
			  ,sizeof(dds->parm[pn]->param_description));
	   jj = dds->field_id_num[pn] =
		  dd_return_id_num(dds->parm[pn]->parameter_name);
	   dgi->parm_type[pn] = 
		  jj == DBZ_ID_NUM || jj == DZ_ID_NUM ? OTHER : VELOCITY;
	   bc += gdsos;
	}
	else if(strncmp(dgi->in_next_block, "COMM", 4) == 0 ) {
# ifdef DEBUG
           printf("#--> Find COMM\n");
# endif
	    dd_next_comment((struct comment_d *)dgi->in_next_block
			    , gottaSwap);
	}
	dgi->buf_bytes_left -= gdsos;
	dgi->file_byte_count += gdsos;
	dgi->in_next_block += gdsos;
# ifdef DEBUG
           printf("====== buf_bytes_left=%d  file_byte_count=%d  in_next_block=%s\n",
                  dgi->buf_bytes_left,dgi->file_byte_count,dgi->in_next_block);
# endif
    }
    if(broken) {
	  /* oops! This is probably a garbaged file */
	  printf("#ERROR: Current sweep file is garbage!\n");
	  printf("        Last time from previous sweep is %s\n"
		         ,dts_print(d_unstamp_time(dgi->dds->dts)));
	  exit(1);
    }

    /* apply range correction to the corrected celv */
    for(ii=0; ii < dds->celv->number_cells; ii++) {
	dds->celvc->dist_cells[ii] += dds->cfac->range_delay_corr;
    }
    dd_set_uniform_cells(dgi->dds);
    dd_absorb_rotang_info(dgi, gottaSwap);
    dd_absorb_seds(dgi, gottaSwap);

    return(0);
}
/* c------------------------------------------------------------------------ */

int dd_absorb_ray_info(dgi)
    DGI_PTR dgi;
{
    /*
     * Keep sucking up input until all headers (if applicable )
     * and a complete ray have been read in.
     */
    int ii, mm, n, nc, len, ryib_flag=NO, num_rdats=0, mark, *deep6=0;
    int num_cells, final_bad_count, bc=0, pn, ncopy;
    int broken = NO, gottaSwap=NO, ds;
    float rcp_scale, bias;
    char *a, *b;
    char str[128];
    double dorade_time_stamp();
    static int count=0, doodah=110, rec_count=0, loop_count = 0;
    static int rec_trip=12, gd_ptr=-1, disk_offset= -1;
    static char *gd_list[22], *last_block=NULL, *last_last_block=NULL;
    struct dd_input_filters *difs, *dd_return_difs_ptr();
    struct paramdata_d *rdat;
    struct qparamdata_d *qdat;
    struct ray_i *ryib;
    struct platform_i *asib;
    struct generic_descriptor *gd;
    DDS_PTR dds=dgi->dds;
    struct radar_angles *ra=dds->ra;
    DD_TIME *dts=dds->dts, *d_unstamp_time();
    struct dd_input_sweepfiles_v3 *dis, *dd_return_sweepfile_structs_v3();
    unsigned long gdsos;

    count++;
    if( count >= doodah ) {
	mark = MAX_REC_SIZE;
    }
    difs = dd_return_difs_ptr();
    dis = dd_return_sweepfile_structs_v3();


    while(1) {
       if( ++loop_count >= doodah ) {
	  mark = MAX_REC_SIZE;
       }
	gd = (struct generic_descriptor *)dgi->in_next_block;

	if(dgi->buf_bytes_left >= sizeof(struct generic_descriptor)) {
	   gdsos = (unsigned long)gd->sizeof_struct;

	   if(gdsos > MAX_REC_SIZE) {
	      gottaSwap = YES;
	      swack4(&gd->sizeof_struct, &gdsos);
	   }
	}
	else { gdsos = 0; }

	if( num_rdats == dgi->source_num_parms ) {
	    break;
	}
	if(dgi->buf_bytes_left <= 0 ||
	   sizeof(struct generic_descriptor) > dgi->buf_bytes_left ||
	   gdsos > dgi->buf_bytes_left ) {
	    /* move to top of buffer and refill it */
	    rec_count++;
	    if(rec_count >= rec_trip) {
		mark = 0;
	    }
	    if(dgi->buf_bytes_left > 0 )
		  memcpy(dgi->in_buf, dgi->in_next_block
			 , dgi->buf_bytes_left);
	    else if( dgi->buf_bytes_left < 0 ) {
		printf("bytes left < 0 count: %d rec_num: %d\n"
		       , count, rec_count);
	        exit(1);
	    }
	    dgi->in_next_block = dgi->in_buf;
	    disk_offset = lseek(dgi->in_swp_fid, 0L, 1L); 

	    if((len=read(dgi->in_swp_fid, dgi->in_buf+dgi->buf_bytes_left
			 , MAX_READ -dgi->buf_bytes_left )) < 1 ) {
	       printf("Trying to read past end of data...stat= %d loop: %d\n"
		      , len, loop_count );
	        exit(1);
	    }
	    else if((dgi->buf_bytes_left+=len) < gdsos ) {
		printf("File appears to be truncated...stat= %d loop: %d\n"
		       , len, loop_count );
		exit(1);
	    }
	    dis->MB += BYTES_TO_MB(len);
	    dis->rec_count++;
	    gd = (struct generic_descriptor *)dgi->in_next_block;
	    gdsos = (unsigned long)gd->sizeof_struct;
# ifdef DEBUG
            printf("+++++++ Read record count=%d (%f MB) +++++++\n",
            dis->rec_count,dis->MB);
# endif
	    if(gdsos > MAX_REC_SIZE) {
	       gottaSwap = YES;
	       swack4(&gd->sizeof_struct, &gdsos);
	    }
	}

	if(gdsos < sizeof(struct generic_descriptor) || gdsos > MAX_REC_SIZE) {
	   /* This is probably a garbaged file */
	   printf("Sweep file containing time %s is garbage! loop: %d\n"
		  , dts_print(d_unstamp_time(dgi->dds->dts)), loop_count);
	   exit(1);
	}
	gd_ptr = INC_NDX(gd_ptr, 22);
	gd_list[gd_ptr] = dgi->in_next_block;

	if(strncmp(dgi->in_next_block,"RDAT",4) == 0 ||
	   strncmp(dgi->in_next_block,"QDAT",4) == 0) {
	    qdat = (struct qparamdata_d *)dgi->in_next_block;
	    pn = num_rdats++;
	    final_bad_count = 0;
	    if(*dgi->in_next_block == 'R') {
	       ncopy = sizeof(struct paramdata_d);   /* ncopy=16 byte */
            } 
            else {
               ncopy = dds->parm[pn]->offset_to_data;
            }
	    ds = dd_datum_size((int)dds->parm[pn]->binary_format);

            if(!dds->qdat[pn]){
                dds->qdat[pn] = (struct qparamdata_d *)malloc(sizeof(struct qparamdata_d));
            }
	    if(gottaSwap) {
	        ddin_crack_qdat(dgi->in_next_block, dds->qdat[pn], (int)0);
	    }
	    else {
	        memcpy((char *)dds->qdat[pn], dgi->in_next_block, ncopy);
	    }

	    if(dds->radd->data_compress == NO_COMPRESSION) {
		a = dgi->in_next_block + ncopy;
		mm = gdsos - ncopy;
		if(gottaSwap && ds > 1) {
		   if(ds == sizeof(short)) {
		      swack_short(a, dds->qdat_ptrs[pn], mm/sizeof(short));
		   }
		   else {	/* punt! */
		   }
		}
		else {
		   memcpy(dds->qdat_ptrs[pn], a, mm);
		}
	    }
            else {
		a = dgi->in_next_block + ncopy;
		b = dds->qdat_ptrs[pn];
		if(gottaSwap) {
		  n = dd_hrd16LE_uncompressx  /* <USE for NCAR SPOL bistatic data> */
		     ((short *)a, (short *)b, (int)dds->parm[pn]->bad_data
		      , &final_bad_count, dds->celv->number_cells);
		}
		else {
		   n = dd_hrd16_uncompressx
		     ((short *)a, (short *)b, (int)dds->parm[pn]->bad_data
		      , &final_bad_count, dds->celv->number_cells);
		}
		mm = n * dd_datum_size((int)dds->parm[pn]->binary_format);
	    }
	    dds->qdat[pn]->pdata_length = ncopy + dds->sizeof_qdat[pn];

	    /* make the clip gate indicate max last good gate for this ray */
	    if(final_bad_count) 
		  if((nc=num_cells-final_bad_count) > dgi->clip_gate) 
			dgi->clip_gate = nc;
	    bc += gdsos;
	}
	else if(strncmp(dgi->in_next_block,"RYIB",4)==0 && ryib_flag ) {
	    /* it'll probably never get here since its counting "rdat"s */
	    printf("RDAT count off! loop: %d\n", loop_count);
            exit(1);
	}
	else if(strncmp(dgi->in_next_block,"RYIB",4)==0 ) {
	    if(!dds->ryib){
               dds->ryib = (struct ray_i *)malloc(sizeof(struct ray_i));
            }
	    dgi->source_ray_num++;
	    dgi->beam_count++;
	    dgi->ray_num++;
	    dis->ray_count++;
	    if(gottaSwap) {
	       ddin_crack_ryib(dgi->in_next_block, dds->ryib, (int)0);
	    }
	    else {
	       memcpy((char *)dds->ryib, dgi->in_next_block
		      , sizeof(struct ray_i));
	    }
	    ryib_flag = YES;
	    bc += gdsos;
	    num_cells = dds->celv->number_cells;
	    dgi->clip_gate = 0;
# ifdef DEBUG
            printf("#--> Find RYIB: alloc sizeof_ray=%d\n",sizeof(struct ray_i));
            print_ray_info(dgi);
# endif
	}
	else if(strncmp(dgi->in_next_block,"ASIB",4)==0) {
	   if(gottaSwap) {
	      ddin_crack_asib(dgi->in_next_block, dds->asib, (int)0);
	   }
	   else {
	      memcpy((char *)dds->asib, dgi->in_next_block
		     , sizeof(struct platform_i));
	   }
	    bc += gdsos;
	}
	else if(strncmp(dgi->in_next_block, "COMM", 4) == 0 ) {
	    dd_next_comment((struct comment_d *)dgi->in_next_block
			    , gottaSwap);
	}
	else if(strncmp(dgi->in_next_block,"NULL",4)==0) {
	    /* it'll probably never get here since its counting rays
	     * and rdats
	     */
# ifdef DEBUG
	    printf("Read the NULL record...ray count off!\n");
            printf("swib:%d source:%d beams:%d rnum:%d rcount:%d loop: %d\n"
		   , dgi->dds->swib->num_rays
		   , dgi->source_ray_num
		   , dgi->beam_count
		   , dgi->ray_num
		   , dis->ray_count
		   , loop_count
		   );
# endif
            return(1);

	}

	else {			/* ignore this descriptor */
	    mark = 0;
	    bc += gdsos;
	}

	dgi->buf_bytes_left -= gdsos;
	dgi->file_byte_count += gdsos;
	dgi->in_next_block += gdsos;
    } /* end of while(1) due to break by (num_rdats == dgi->source_num_parms) */

    dgi->time = dts->time_stamp =
	  dorade_time_stamp(dds->vold, dds->ryib, dts);
    d_unstamp_time(dts);

    if( dgi->dds->radd->scan_mode  == AIR ||
        dgi->dds->radd->radar_type == AIR_LF ||
	dgi->dds->radd->radar_type == AIR_NOSE )
    {
	 dd_radar_angles( dds->asib, dds->cfac, ra, dgi );
    }
    if(difs->altitude_truncations) {
	dgi->clip_gate = dd_clip_gate
	      (dgi, dds->ra->elevation, dds->asib->altitude_msl
	       , difs->altitude_limits->lower
	       , difs->altitude_limits->upper);
    }
    else if(!dgi->clip_gate)
	  dgi->clip_gate = dds->celv->number_cells-1;

    if(!dgi->time0 ) {	/* first time processed */
	dgi->time0 = dgi->time;
    }

    return(bc);
}
/* c------------------------------------------------------------------------ */

int dd_absorb_rotang_info(dgi, gottaSwap)
    DGI_PTR dgi;
    int gottaSwap;
{
    int i, jj, n;
    struct rot_ang_table *rat;
    char *c, *swap_buf;
    long disk_offset;
    DDS_PTR dds=dgi->dds;

    disk_offset = lseek(dgi->in_swp_fid, 0L, 1L); /* remember the disk offset */
    /*
     * now position to the rotation angle table
     * and absorb it
     */
    for(i=0; i < MAX_KEYS; i++) {
	if(dds->sswb->key_table[i].type == KEYED_BY_ROT_ANG)
	      break;
    }
    n = lseek(dgi->in_swp_fid, (long)dds->sswb->key_table[i].offset, 0L);
    if(dgi->source_rat) {
	free(dgi->source_rat);
    }
    c = (char *)malloc(dds->sswb->key_table[i].size);
    rat = dgi->source_rat = (struct rot_ang_table *)c;
    if(gottaSwap) {
       swap_buf = (char *)malloc(dds->sswb->key_table[i].size);
       n = read(dgi->in_swp_fid, swap_buf, dds->sswb->key_table[i].size);
       ddin_crack_rktb(swap_buf, rat, (int)0);
       jj = rat->angle_table_offset;
       swack_long(swap_buf + jj, (char *)rat + jj, rat->ndx_que_size);
       jj = rat->first_key_offset;
       swack_long(swap_buf + jj, (char *)rat + jj, rat->num_rays * 3);
       free(swap_buf);
    }
    else {
       n = read(dgi->in_swp_fid, c, dds->sswb->key_table[i].size);
    }
    /* now put it back
     */
    disk_offset = lseek(dgi->in_swp_fid, (long)disk_offset, 0L);
    return(0);
}
/* c------------------------------------------------------------------------ */

int dd_absorb_seds(dgi, gottaSwap)
    DGI_PTR dgi;
    int gottaSwap;
{
    /* absorb the editing summary */
    int ii, nn;
    long disk_offset, gdsos;
    struct generic_descriptor *gd;
    DDS_PTR dds=dgi->dds;

    /* remember the disk offset
     */
    disk_offset = lseek(dgi->in_swp_fid, 0L, 1L);
    /*
     * now position to the solo edit summary
     * and absorb it
     */
    for(ii=0; ii < dds->sswb->num_key_tables; ii++) {
	if(dds->sswb->key_table[ii].type == SOLO_EDIT_SUMMARY)
	      break;
    }
    if(ii == dds->sswb->num_key_tables) {
	/* no seds */
	return(0);
    }
    if(dgi->seds) free(dgi->seds);

    dgi->sizeof_seds = dds->sswb->key_table[ii].size;

    if((dgi->sizeof_seds = dds->sswb->key_table[ii].size) >= K64) {
       dgi->seds = NULL;
       dgi->sizeof_seds = 0;
       return(1);
    }
    else if(!(dgi->seds = (char *)malloc(dgi->sizeof_seds))) {
       dgi->seds = NULL;
       dgi->sizeof_seds = 0;
       return(1);
    }
    nn = lseek(dgi->in_swp_fid, (long)dds->sswb->key_table[ii].offset, 0L);
    nn = read(dgi->in_swp_fid, dgi->seds, dgi->sizeof_seds);

    if(gottaSwap) {
       gd = (struct generic_descriptor *)dgi->seds;
       swack4(&gd->sizeof_struct, &gdsos);
       gd->sizeof_struct = gdsos;
    }
    /*
     * now move disk back to where it was
     */
    disk_offset = lseek(dgi->in_swp_fid, (long)disk_offset, 0L);
    return(0);
}

/* c------------------------------------------------------------------------ */

struct dd_input_sweepfiles_v3 *
dd_return_sweepfile_structs_v3()
{
    if(!dis_v3) {
	dis_v3 = (struct dd_input_sweepfiles_v3 *)
	      malloc(sizeof(struct dd_input_sweepfiles_v3));
	memset(dis_v3, 0, sizeof(struct dd_input_sweepfiles_v3));
	dis_v3->usi_top = (struct unique_sweepfile_info_v3 *)
	      malloc(sizeof(struct unique_sweepfile_info_v3));
	memset(dis_v3->usi_top, 0, sizeof(struct unique_sweepfile_info_v3));

    }
    return(dis_v3);
}

/*=========================================================================
/*     below functions are copied from other source code
/*     "dd_return_difs_ptr"       from dda_common.c
/*     "dd_next_comment"          from dda_common.c
/*     "dd_hrd16_uncompressx"     from dorade_share.c
/*     "dd_hrd16LE_uncompressx"   from dorade_share.c
/*     "dts_print"                from dorade_share.c
/*     "str_terminate"            from dorade_share.c
/*     "dorade_time_stamp"        from dorade_share.c
/*========================================================================= */
/* c------------------------------------------------------------------------ */

struct dd_input_filters *
dd_return_difs_ptr()
{
    if(!keep_difs) {
	keep_difs = (struct dd_input_filters *)
	      malloc(sizeof(struct dd_input_filters));
	memset((char *)keep_difs, 0, sizeof(struct dd_input_filters));

    }
    return(keep_difs);
}

/* c------------------------------------------------------------------------ */

struct dd_comm_info *
dd_next_comment(comm, little_endian)
  struct comment_d *comm;
{
    /* returns a pointer to the next comment info struct
     * this routine expects the calling routine to malloc space for
     * the actual comment, set the pointer and copy the comment into it
     */
    int nn = 0, cdl=0;
    struct dd_comm_info *this=first_comment;

    if(!comm)
          return(NULL);

    if(little_endian) {
       swack4((char *)comm +4, &cdl);
    }
    else {
       cdl = comm->comment_des_length;
    }

    if(cdl < 8 || cdl > MAX_REC_SIZE) {
        return(NULL);
    }

    if(first_comment)
          nn = first_comment->num_comments;

    for(; nn-- ; this = this->next);

    if(!this) {
        this = (struct dd_comm_info *)malloc(sizeof(struct dd_comm_info));
        memset((char *)this, 0, sizeof(struct dd_comm_info));
        if(!first_comment) first_comment = this;
    }
    if(!this->comm ||
      this->comm->comment_des_length < cdl) {
        if(this->comm) free(this->comm);
        this->comm = (struct comment_d *)malloc(cdl);
        memset(this->comm, 0, cdl);
    }
    memcpy(this->comm, comm, cdl);
    this->comm->comment_des_length = cdl;

    if(first_comment->num_comments++)
          first_comment->last->next = this;
    first_comment->last = this; /* always points to last comment */
    return(this);
}
/* c------------------------------------------------------------------------ */

int dd_hrd16_uncompressx( ss, dd, flag, empty_run, wmax )
  short *ss, *dd;
  int flag, *empty_run, wmax;
{
    /*
     * routine to unpacks actual data assuming MIT/HRD compression where:
     * ss points to the first 16-bit run-length code for the compressed data
     * dd points to the destination for the unpacked data
     * flag is the missing data flag for this dataset that is inserted
     *     to fill runs of missing data.
     * empty_run pointer into which the number of missing 16-bit words
     *    can be stored. This will be 0 if the last run contained data.
     # wmax indicate the maximum number of 16-bit words the routine can
     *    unpack. This should stop runaways.
     */
    int i, j, k, n, mark, wcount=0;

    while(*ss != 1) {		/* 1 is an end of compression flag */
	n = *ss & 0x7fff;	/* nab the 16-bit word count */
	if(wcount+n > wmax) {
	    printf("Uncompress failure %d %d %d\n"
		   , wcount, n, wmax);
	    mark = 0;
	    break;
	}
	else {
	    wcount += n;		/* keep a running tally */
	}
	if( *ss & 0x8000 ) {	/* high order bit set implies data! */
	    *empty_run = 0;
	    ss++;
	    for(; n--;) {
		*dd++ = *ss++;
	    }
	}	
	else {			/* otherwise fill with flags */
	    *empty_run = n;
	    ss++;
	    for(; n--;) {
		*dd++ = flag;
	    }
	}
    }
    return(wcount);
}
/* c------------------------------------------------------------------------ */

int dd_hrd16LE_uncompressx( ss, dd, flag, empty_run, wmax )
  unsigned short *ss, *dd;
  int flag, *empty_run, wmax;
{
    /*
     * routine to unpacks actual data assuming MIT/HRD compression where:
     * ss points to the first 16-bit run-length code for the compressed data
     * dd points to the destination for the unpacked data
     * flag is the missing data flag for this dataset that is inserted
     *     to fill runs of missing data.
     * empty_run pointer into which the number of missing 16-bit words
     *    can be stored. This will be 0 if the last run contained data.
     # wmax indicate the maximum number of 16-bit words the routine can
     *    unpack. This should stop runaways.
     */
    int i, j, k, n, mark, wcount=0;
    unsigned short rlcw;
    unsigned char *aa, *bb;
    
    aa = (unsigned char *)&rlcw;

    for(;;) {	
       bb = (unsigned char *)ss;
       *aa = *(bb+1);
       *(aa+1) = *bb;		/* set run length code word "rlcw" */
       if(rlcw == 1) { break; }	/* 1 is the end of compression flag */

       n = rlcw & 0x7fff;	/* nab the 16-bit word count */
       if(wcount+n > wmax) {
	  printf("Uncompress failure %d %d %d\n"
		 , wcount, n, wmax);
	  mark = 0;
	  break;
       }
       else {
	  wcount += n;		/* keep a running tally */
       }
       if( rlcw & 0x8000 ) {	/* high order bit set implies data! */
	  *empty_run = 0;
	  ss++;
	  swack_short(ss, dd, n);
	  ss += n;
	  dd += n;
	  
       }	
       else {			/* otherwise fill with flags */
	  *empty_run = n;
	  ss++;
	  for(; n--;) {
	     *dd++ = flag;
	  }
       }
    }
    return(wcount);
}
/* c------------------------------------------------------------------------ */

char *dts_print(dts)
  DD_TIME *dts;
{
    static char str[32];

    sprintf(str, "%02d/%02d/%02d %02d:%02d:%02d.%03d"
           , dts->month
           , dts->day
           , dts->year-1900
           , dts->hour
           , dts->minute
           , dts->second
           , dts->millisecond
           );
    return(str);
}

/* c------------------------------------------------------------------------ */
char *str_terminate(dst, srs, n)
  CHAR_PTR srs, dst;
  int n;
{
    int m=n;
    char *a=srs, *c=dst;

    *dst = '\0';

    if(n < 1)
          return(dst);
    /*
     * remove leading blanks
     */
    for(; m && *a == ' '; m--,a++);
    /*
     * copy m char or to the first null
     */
    for(; m-- && *a; *c++ = *a++);
    /*
     * remove trailing blanks
     */
    for(*c='\0'; dst < c && *(c-1) == ' '; *(--c)='\0');

    return(dst);
}
/* c------------------------------------------------------------------------ */

double dorade_time_stamp( vold, ryib, dts )
  struct volume_d *vold;
  struct ray_i *ryib;
  DD_TIME *dts;
{

    double d, d_time_stamp();

/*  original:  dts->year = vold->year > 1900 ? vold->year : vold->year +1900; */
    if(vold->year > 1900){
       dts->year = vold->year;
    }else if(vold->year > 50){
       dts->year = vold->year + 1900;
# ifdef DEBUG
       printf("### To avoid YK2 Problem, if(vold->year>50) vold->year+1900 ####\n"); 
# endif
    }else{
       dts->year = vold->year + 2000;
    }
    dts->month = dts->day = 0;
    dts->day_seconds = (ryib->julian_day-1)*SECONDS_PER_DAY
	  +D_SECS(ryib->hour, ryib->minute, ryib->second
		  , ryib->millisecond);
    d = d_time_stamp(dts);
	  
    return(d);
 }
